import abc
import base64
import collections
import enum
import json
import logging
import os.path
import pathlib
import re
import subprocess
import textwrap
import time

from typing import Dict

import asn1crypto.cms

from . import plugins
from . import settings
from .wrap import livestatus

log = logging.getLogger(__name__)

_PLUGIN_EXIT_CODES = {
    0: "OK",
    1: "WARNING",
    2: "CRITICAL",
    3: "UNKNOWN",
}


class CheckResult(enum.Enum):
    passed = 1
    failed = 2


class CheckType(abc.ABC):

    name = ""

    def __init__(self, config):
        self._conf = config
        self.errors = []

    @abc.abstractmethod
    def run(self):
        pass

    @abc.abstractmethod
    def result(self) -> CheckResult:
        pass

    @abc.abstractmethod
    def report(self):
        pass

    def __str__(self):
        return self.name

    @classmethod
    def get_enabled_checks(cls, config):
        enabled = []
        if config.checks is None:
            enabled = [c(config) for c in cls.__subclasses__()]
        else:
            enabled = [
                c(config) for c in cls.__subclasses__() if c.name in config.checks
            ]
        log.info("All checks selected: %s", [str(c) for c in enabled])
        return enabled

    @classmethod
    def all_check_names(cls):
        return [c.name for c in cls.__subclasses__()]

    @staticmethod
    def heading(text):
        return f"*** {text}"


class Livestatus:
    def __init__(self, use_lmd=True):
        socket = "/opt/monitor/var/rw/live{}".format("_tmp" if not use_lmd else "")
        log.debug("Use Livestatus socket %s", socket)
        self.conn = livestatus.SingleSiteConnection("unix:" + socket)
        self.errors = []

    def get_check_commands(self) -> Dict[str, plugins.CheckPlugin]:
        """Get list of configured OP5 check commands.

        OP5 check commands are those with command line matching;
            * Path that starts with one of
                - $USER1$
                - $USER2$
                - /opt/plugins/
            * except paths in `custom/` subdirectory of above.
        """
        query = textwrap.dedent(
            """\
            GET commands
            Columns: name line
            """
        )
        log.debug("Livestatus query: %r", query)
        data = self.conn.query_table(query)
        log.debug("GET commands: got %s commands", len(data))
        cmds = {}
        for cmd_name, cmd_line in data:
            try:
                command = cmd_line.split(maxsplit=1)[0]
                match = re.match(r"(?:\$USER[12]\$|/opt/plugins)/(custom/)?", command)
                if not match:
                    log.debug("Skip %s - not a plugin", cmd_name)
                    continue
                if match.group(1):
                    log.info("Skip %s - uses custom plugin %r", cmd_name, command)
                    continue
                cmds[cmd_name] = plugins.CheckPlugin(os.path.basename(command), command)
            except Exception as exc:
                log.warning("failed to parse %r: %s", (cmd_name, cmd_line), exc)
        return cmds

    def services(self, page_size, max_pages):
        self.conn.set_limit(page_size)
        tmpl = textwrap.dedent(
            """\
            GET services
            Columns: host_name description check_command
            Sort: host_name asc
            Sort: description asc
            Offset: {offset}
            """
        )
        for page in range(max_pages):  # limit number of queries to avoid running amok
            query = tmpl.format(offset=page * page_size)
            log.debug("Livestatus query: %r", query)
            data = self.conn.query_table(query)
            if not data:
                break
            yield from data
        else:
            self._log_error(
                "Stop returning data after reaching max number of Livestatus requests "
                "(%d). You can try increasing the limit using '--max-pages N'.",
                max_pages,
            )

    def _log_error(self, *args, **kwargs):
        log.error(*args, **kwargs)
        self.errors.append(args[0] % args[1:])


class CheckPlugins(CheckType):
    """Check that the plugins currently in use are available on the next
    version of OP5 Monitor."""

    name = "plugins"
    report_dir = pathlib.Path("/var/tmp")

    def __init__(self, *args):
        super().__init__(*args)
        self.problem_plugins = collections.defaultdict(dict)
        self.plugin_notes = collections.defaultdict(dict)
        self.count_plugins = collections.Counter()
        self.count_commands = collections.Counter()

    def run(self):
        log.info("Running check '%s'", self.name)
        print("Checking plugin usage of all services in the configuration.")
        print("This may take a while, depending on the size of your configuration...")
        available_plugins = plugins.PluginSet()
        all_plugin_notes = plugins.get_plugin_notes()
        log.debug("Monitor version 9 plugins: %s", sorted(available_plugins))
        ls = Livestatus(self._conf.use_lmd)
        op5_check_commands = ls.get_check_commands()
        log.info("Current configuration has %d check commands", len(op5_check_commands))
        seen = set()
        for host_name, description, cmdline in ls.services(
            self._conf.page_size, self._conf.max_pages
        ):
            cmd_name = cmdline.split("!", maxsplit=1)[0]
            self.count_commands[cmd_name] += 1
            try:
                plugin = op5_check_commands[cmd_name]
            except KeyError:
                # e.g. services that use non-op5 plugins
                log.debug(
                    "Skip service '%s;%s' (command: %s)",
                    host_name,
                    description,
                    cmd_name,
                )
                continue
            self.count_plugins[plugin.path] += 1

            if plugin.name in all_plugin_notes and plugin.name not in self.plugin_notes:
                self.plugin_notes[plugin.name] = all_plugin_notes[plugin.name]

            if plugin not in available_plugins:
                self.problem_plugins[plugin.name].setdefault(cmd_name, []).append(
                    {"host_name": host_name, "description": description}
                )
                if not self._conf.debug and cmd_name in seen:
                    # Log only once per plugin
                    continue
                log.info(
                    "Plugin '%s' used by check command '%s' of '%s;%s' is not marked "
                    "as included in target version. Subsequent messages for %s will "
                    "not be logged.",
                    plugin.name,
                    cmd_name,
                    host_name,
                    description,
                    plugin.name,
                )
                seen.add(cmd_name)
        self.errors += ls.errors
        log.info("Finished check '%s'", self.name)

    def result(self):
        if self.problem_plugins or self.errors:
            return CheckResult.failed
        return CheckResult.passed

    def _dump_json_report(self):
        self.report_file = self.report_dir / "op5-plugin-report-{}.json".format(
            time.strftime("%Y-%m-%d_%H%M%S")
        )
        report = {
            "problem_plugins": self.problem_plugins,
            "plugin_notes": self.plugin_notes,
            "plugin_service_count": self.count_plugins,
            "check_command_service_count": self.count_commands,
        }
        with self.report_file.open("w") as fp:
            json.dump(report, fp, indent=4)
        log.info("Plugin report written to %s", self.report_file)

    def report(self):
        self._dump_json_report()
        print(self.heading("Current OP5 Monitor configuration"))
        print("Number of services: {}".format(sum(self.count_commands.values())))
        print(
            "Number of check commands in use by services: {}".format(
                len(self.count_commands)
            )
        )
        print("Number of plugins used: {}".format(len(self.count_plugins)))
        print("Plugins used:")
        for plugin in sorted(self.count_plugins.keys()):
            print("  {}".format(plugin))

        print(self.heading("Plugin usage report"))
        if self.problem_plugins:
            print("Problems found:")
        for plugin, check_commands in self.problem_plugins.items():
            print(
                "Plugin '{}' used by {} check commands and {} service checks:".format(
                    plugin,
                    len(check_commands),
                    sum(len(x) for x in check_commands.values()),
                )
            )
            # If there is a specific note about this plugin, print it here and
            # remove it from notes collection, as any remaining notes are printed
            # later.
            note = self.plugin_notes.pop(
                plugin, f"'{plugin}' is currently not available in the target version."
            )
            if note:
                lines = textwrap.wrap(
                    note,
                    width=70,
                    initial_indent="  > ",
                    subsequent_indent="    ",
                )
                print("\n".join(lines))
            print("")

        # If there are notes left, not printed above, display them as
        # recommendations about used plugins.
        if self.plugin_notes:
            print("Specific plugin recommendations:")
            for plugin_name, note in self.plugin_notes.items():
                print(plugin_name)
                lines = textwrap.wrap(
                    note,
                    width=70,
                    initial_indent="  > ",
                    subsequent_indent="    ",
                )
                print("\n".join(lines))

        if self.problem_plugins:
            print(
                "\nAll check commands and services that use the deprecated plugins are "
                "listed in\n{}".format(self.report_file)
            )
        elif not self.plugin_notes:
            print("No problems found with the plugins currently in use.")


class CheckLicense(CheckType):
    """Check that installed OP5 Monitor license is;

    * valid for > 60 days
    * not too old to work with next major version (creation date and license version)
    """

    name = "license"
    license_path = "/etc/op5license/op5license.lic"
    license_ca_common_name = "OP5 License Root"

    def __init__(self, *args):
        super().__init__(*args)
        self._retcode = None
        self._stdout = "n/a"
        self._license_ok = None
        self._trial_license = False

    warning_days = 60
    critical_days = 30

    def run(self):
        log.info("Running check '%s'", self.name)
        if not os.path.exists(self.license_path):
            log.info(
                "License file %s doesn't exist, skip running license checks",
                self.license_path,
            )
            self._trial_license = True
            return
        self._run_check_op5_license()
        try:
            self._license_ok = self._is_v9_compatible()
        except Exception as e:
            log.warning("Cannot determine license compatibility: %s", e)
            self.errors.append(str(e))
            self._license_ok = False
        log.info("Finished check '%s'", self.name)

    def _run_check_op5_license(self):
        try:
            p = subprocess.run(
                [
                    "/opt/plugins/check_op5_license",
                    "--type=d",
                    "--warning",
                    str(self.warning_days),
                    "--critical",
                    str(self.critical_days),
                ],
                universal_newlines=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                timeout=10,
            )
        except Exception as e:
            log.info("Got exception", exc_info=True)
            self.errors.append(str(e))
        else:
            self._retcode = p.returncode
            self._stdout = p.stdout
            if self._retcode > 3:
                self.errors.append(self._stdout)

    def _is_v9_compatible(self):
        path = pathlib.Path(self.license_path)
        with path.open("rb") as f:
            lines = f.readlines()
        first = lines.pop(0).strip()
        last = lines.pop().strip()
        cms_bytes = base64.standard_b64decode(b"".join(lines))
        if not (
            first == b"-----BEGIN OP5 LICENSE BLOCK-----"
            and last == b"-----END OP5 LICENSE BLOCK-----"
        ):
            log.info("License file format is too old")
            return False

        # Load license CMS data and verify that signing certificate issuer (the
        # root CA) is the new one that is required on Monitor 9.
        info = asn1crypto.cms.ContentInfo.load(cms_bytes)
        if info["content_type"].native != "signed_data":
            raise ValueError("Cannot parse license data")
        signed_data = info["content"]
        signer_cert = signed_data["certificates"][0].chosen
        log.debug("Signed by cert serial_number: %s", signer_cert.serial_number)
        tbs_cert = signer_cert["tbs_certificate"]
        log.debug("tbs_certificate: %s", tbs_cert.native)
        if tbs_cert["issuer"].native["common_name"] == self.license_ca_common_name:
            return True
        return False

    def result(self):
        if self._trial_license:
            return CheckResult.passed
        if not self.errors and self._retcode == 0 and self._license_ok:
            return CheckResult.passed
        return CheckResult.failed

    def report(self):
        if self._trial_license:
            print("No license file installed - currently running trial license.")
            return
        print(self.heading("check_op5_license"))
        print("Plugin output: {}".format(self._stdout.split("|")[0]))
        print(
            "Plugin check result: {}".format(
                _PLUGIN_EXIT_CODES.get(self._retcode, "error")
            )
        )
        if not self._license_ok:
            print(self.heading("License file OP5 Monitor 9 compatibility"))
            print(
                "Installed license file version is too old. "
                "Please request a new license file from {}.".format(
                    settings.SUPPORT_NAME
                )
            )


class CheckVersion(CheckType):
    """Check that installed OP5 Monitor version is recent enough."""

    name = "op5-version"
    min_version = ("8.3.8", (2022, 2, 1))
    package_name = "op5-monitor"
    _version_is_ok = None

    def run(self):
        log.info("Running check '%s'", self.name)
        command = ["rpm", "-q", self.package_name, "--queryformat", "%{version}"]
        log.debug("Run command: %s", command)
        stdout = None
        returncode = None
        try:
            p = subprocess.run(
                command,
                universal_newlines=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                timeout=3,
            )
        except Exception as e:
            log.info("Got exception", exc_info=True)
            self.errors.append(str(e))
        else:
            returncode = p.returncode
            stdout = p.stdout
            if returncode != 0:
                self.errors.append(stdout)
        log.debug("Return code: %s, stdout: %r", returncode, stdout)
        if returncode == 0:
            try:
                self._version_is_ok = self._check_version(stdout)
            except Exception as e:
                log.warning("Failed to check %s version: %s", self.package_name, e)
                self.errors.append(str(e))
        log.info("Finished check '%s'", self.name)

    @classmethod
    def _check_version(cls, stdout):
        version = tuple(int(n) for n in stdout.strip().split("."))
        return version >= cls.min_version[1]

    def result(self):
        return (
            CheckResult.passed
            if not self.errors and self._version_is_ok is True
            else CheckResult.failed
        )

    def report(self):
        if self._version_is_ok is True or self.errors:
            return
        message = textwrap.dedent(
            """\
            To use the automated configuration migration tools op5-backup/op5-restore,
            OP5 Monitor needs to be updated to at least version {} to support migration
            to OP5 Monitor version 9.
        """
        ).format(self.min_version[0])
        print(message)
