import argparse
import logging
import os
import sys
import tempfile

from . import check
from . import settings

log = logging.getLogger(__package__)
config = None


class OP5Error(Exception):
    """Installer errors.

    Args:
        message: error message.
        *args: arguments for %s formatting of message.
        retcode: attach a return code to exit the program with.
    """

    _default_retcode = 1

    def __init__(self, message, *args, retcode=None):
        self._message = message
        self._args = args
        self.retcode = retcode if retcode is not None else self._default_retcode
        super().__init__(message, *args)

    def __str__(self):
        return self._message % self._args


class _AppConfig:

    _default_loglevel = logging.WARNING

    def __init__(
        self,
        *,
        checks: list,
        verbose: int,
        max_pages: int,
        page_size: int,
        use_lmd: bool,
        logger=None,
        debug=False,
    ):
        self._verbosity = verbose
        self.debug = debug
        if debug:
            self._verbosity = 3
        self._init_logging(logger)
        self.max_pages = max_pages
        self.page_size = page_size
        self.use_lmd = use_lmd
        self.checks = checks

    def __del__(self):
        try:
            self.log_file.close()
        except Exception:
            pass

    def _init_logging(self, logger):
        console = logging.StreamHandler()
        log_level = max(self._default_loglevel - self._verbosity * 10, logging.DEBUG)
        self.debuglog_enabled = log_level == logging.DEBUG
        debug_formatter = logging.Formatter(
            "%(asctime)s %(name)s %(filename)s:%(lineno)s "
            "%(levelname)s: %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        default_formatter = logging.Formatter(
            "%(asctime)s %(levelname)s: %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
        )
        console.setLevel(log_level)
        console.setFormatter(debug_formatter if self.debug else default_formatter)

        log_fd, self.log_filename = tempfile.mkstemp(
            suffix=".log", prefix="op5-", dir="/var/tmp"
        )
        self.log_file = os.fdopen(log_fd, "w")
        fh = logging.StreamHandler(self.log_file)
        fh.setFormatter(debug_formatter)

        self.logger = logger if logger else logging.getLogger(__name__)
        self.logger.setLevel(logging.DEBUG)
        self.logger.addHandler(console)
        self.logger.addHandler(fh)


def main(argv):
    parser = argparse.ArgumentParser(
        description="Check OP5 Monitor major version compatibility",
    )
    parser.add_argument(
        "--check",
        action="append",
        metavar="NAME",
        dest="checks",
        help="Enable specific check. Can be used multiple times. "
        "(Available checks: {})".format(", ".join(check.CheckType.all_check_names())),
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase console logging verbosity.",
    )
    parser.add_argument("--debug", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument(
        "--max-pages",
        type=int,
        default=settings.DEFAULT_MAX_PAGES,
        help=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--page-size",
        type=int,
        default=settings.DEFAULT_PAGE_SIZE,
        help=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--no-lmd",
        action="store_false",
        dest="use_lmd",
        help="Use Livestatus directly instead of LMD.",
    )
    args = parser.parse_args(argv)
    global config
    config = _AppConfig(
        checks=args.checks,
        verbose=args.verbose,
        debug=args.debug,
        logger=log,
        max_pages=args.max_pages,
        page_size=args.page_size,
        use_lmd=args.use_lmd,
    )
    log.info("Debug logfile: %s", config.log_filename)

    enabled_checks = check.CheckType.get_enabled_checks(config)
    check_results = {}
    for chk in enabled_checks:
        print("\n==========================================================")
        print("= Running check: {}".format(chk))
        print("==========================================================")
        chk.run()
        check_results[chk.name] = chk.result()

    for chk in enabled_checks:
        print("\n==========================================================")
        print("= Report: {}".format(chk))
        print("==========================================================")
        chk.report()
        print("{} result: {}".format(chk, chk.result().name))
        if chk.errors:
            print("\nErrors:")
            for err in chk.errors:
                print("*", err)

    if len(enabled_checks) != len(check_results):
        log.error(
            "Enabled checks %s, but only check results for %s",
            enabled_checks,
            check_results,
        )

    print("\n==========================================================")
    print("= Summary")
    print("==========================================================")
    print(
        "{}/{} checks completed successfully".format(
            sum(1 for c in check_results.values() if c == check.CheckResult.passed),
            len(enabled_checks),
        )
    )
    failed_checks = [
        n for n, res in check_results.items() if res != check.CheckResult.passed
    ]
    if failed_checks:
        print("Failed checks: {}".format(", ".join(failed_checks)))
        return False
    return True


def cli_main():
    """Called from package entrypoint script."""
    try:
        if not main(sys.argv[1:]):
            sys.exit(1)
    except OP5Error as exc:
        log.error("%s", exc)
        log.debug("installation failed", exc_info=True)
        log.info("Debug logfile: %s", getattr(config, "log_filename", "n/a"))
        sys.exit(exc.retcode)
    except Exception as exc:
        log.info("Got exception", exc_info=True)
        log.critical(
            "Unexpected failure: %s. Contact %s for assistance. "
            "debug log file: %s\n"
            "(running with --verbose can give more details)",
            exc,
            settings.SUPPORT_NAME,
            getattr(config, "log_filename", "n/a"),
        )
        sys.exit(1)
